/*++
/* NAME
/*	qmqpd_peer 3
/* SUMMARY
/*	look up peer name/address information
/* SYNOPSIS
/*	#include "qmqpd.h"
/*
/*	void	qmqpd_peer_init(state)
/*	QMQPD_STATE *state;
/*
/*	void	qmqpd_peer_reset(state)
/*	QMQPD_STATE *state;
/* DESCRIPTION
/*	The qmqpd_peer_init() routine attempts to produce a printable
/*	version of the peer name and address of the specified socket.
/*	Where information is unavailable, the name and/or address
/*	are set to "unknown".
/*
/*	qmqpd_peer_init() updates the following fields:
/* .IP name
/*	The client hostname. An unknown name is represented by the
/*	string "unknown".
/* .IP addr
/*	Printable representation of the client address.
/* .IP namaddr
/*	String of the form: "name[addr]".
/* .PP
/*	qmqpd_peer_reset() releases memory allocate by qmqpd_peer_init().
/* LICENSE
/* .ad
/* .fi
/*	The Secure Mailer license must be distributed with this software.
/* AUTHOR(S)
/*	Wietse Venema
/*	IBM T.J. Watson Research
/*	P.O. Box 704
/*	Yorktown Heights, NY 10598, USA
/*--*/

/* System library. */

#include <sys_defs.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>			/* strerror() */
#include <errno.h>
#include <netdb.h>
#include <string.h>

 /*
  * Older systems don't have h_errno. Even modern systems don't have
  * hstrerror().
  */
#ifdef NO_HERRNO

static int h_errno = TRY_AGAIN;

#define  HSTRERROR(err) "Host not found"

#else

#define  HSTRERROR(err) (\
	err == TRY_AGAIN ? "Host not found, try again" : \
	err == HOST_NOT_FOUND ? "Host not found" : \
	err == NO_DATA ? "Host name has no address" : \
	err == NO_RECOVERY ? "Name server failure" : \
	strerror(errno) \
    )
#endif

#ifdef INET6
#define GAI_STRERROR(error) \
	((error = EAI_SYSTEM) ? gai_strerror(error) : strerror(errno))
#endif

/* Utility library. */

#include <msg.h>
#include <mymalloc.h>
#include <valid_hostname.h>
#include <stringops.h>
#ifdef INET6
#include <inet_addr_list.h>	/* for NI_WITHSCOPEID */
#endif

/* Global library. */

/* Application-specific. */

#include "qmqpd.h"

/* qmqpd_peer_init - initialize peer information */

void    qmqpd_peer_init(QMQPD_STATE *state)
{
    char  *myname = "qmqpd_peer_init";
#ifdef INET6
    struct sockaddr_storage ss;
#else
    struct sockaddr ss;
    struct in_addr *in;
    struct hostent *hp;
#endif
    struct sockaddr *sa;
    SOCKADDR_SIZE len;

    sa = (struct sockaddr *)&ss;
    len = sizeof(ss);

    /*
     * Look up the peer address information.
     */
    if (getpeername(vstream_fileno(state->client), sa, &len) >= 0) {
	errno = 0;
    }

    /*
     * If peer went away, give up.
     */
    if (errno == ECONNRESET || errno == ECONNABORTED) {
	state->name = mystrdup("unknown");
	state->addr = mystrdup("unknown");
    }

    /*
     * Look up and "verify" the client hostname.
     */
    else if (errno == 0 && (sa->sa_family == AF_INET
#ifdef INET6
			    || sa->sa_family == AF_INET6
#endif
             )) {
#ifdef INET6
	char hbuf[NI_MAXHOST];
	char abuf[NI_MAXHOST];
	char rabuf[NI_MAXHOST];
	struct addrinfo hints, *res0 = NULL, *res;
	char *colonp;
#else
	char abuf[sizeof("255.255.255.255") + 1];
	char *hbuf;
#endif
	int error = -1;

#ifdef INET6
	error = getnameinfo(sa, len, abuf, sizeof(abuf), NULL, 0,
			    NI_NUMERICHOST | NI_WITHSCOPEID);
	if (error)
	    msg_fatal("%s: numeric getnameinfo lookup for peer: error %s",
		      myname, GAI_STRERROR(error));
	/*
	 * Convert IPv4-mapped IPv6 address to 'true' IPv4 address
	 * early on. We have no need for the mapped form in logging
	 * or access checks.
	 */
	if (sa->sa_family == AF_INET6
	    && IN6_IS_ADDR_V4MAPPED(&((struct sockaddr_in6 *)sa)->sin6_addr)
	    && (colonp = strrchr(abuf, ':')) != NULL) {
	    if (msg_verbose > 1)
		msg_info("%s: rewriting V4-mapped address \"%s\" to \"%s\"",
			 myname, abuf, colonp + 1);
	    state->addr = mystrdup(colonp + 1);
	} else {
	    state->addr = mystrdup(abuf);
	}

	error = getnameinfo(sa, len, hbuf, sizeof(hbuf), NULL, 0, NI_NAMEREQD);
#else
	in = &((struct sockaddr_in *)sa)->sin_addr;
	inet_ntop(AF_INET, in, abuf, sizeof(abuf));
	state->addr = mystrdup(abuf);
	hbuf = NULL;
	hp = gethostbyaddr((char *)in, sizeof(*in), AF_INET);
	if (hp) {
	    error = 0;
	    hbuf = mystrdup(hp->h_name);
	    state->name = mystrdup("unknown");
	} else {
	    error = 1;
	}
#endif
	if (error) {
	    state->name = mystrdup("unknown");
#ifdef INET6
	    if (error != EAI_NONAME)
		msg_warn("%s: getnameinfo(%s,,,,,,NI_NAMEREQD) error %s",
			 myname, abuf, GAI_STRERROR(error));
#endif
	} else if (!valid_hostname(hbuf, DONT_GRIPE)) {
	    state->name = mystrdup("unknown");
	} else {
	    state->name = mystrdup(hbuf);

	    /*
	     * Reject the hostname if it does not list the peer address.
	     */
#define REJECT_PEER_NAME(state) { \
	myfree(state->name); \
	state->name = mystrdup("unknown"); \
    }

#ifdef INET6
	    memset(&hints, 0, sizeof(hints));
	    hints.ai_family = AF_UNSPEC;
	    hints.ai_socktype = SOCK_STREAM;
	    error = getaddrinfo(state->name, NULL, &hints, &res0);
	    if (error) {
		msg_warn("%s: hostname %s verification failed: %s",
		         state->addr, state->name, GAI_STRERROR(error));
		REJECT_PEER_NAME(state);
	    } else {
		for (res = res0; res; res = res->ai_next) {
		    if (res->ai_family != sa->sa_family)
			continue;
		    error = getnameinfo(res->ai_addr, res->ai_addrlen,
			    rabuf, sizeof(rabuf), NULL, 0,
			    NI_NUMERICHOST | NI_WITHSCOPEID);
		    if (error) {
			msg_warn("%s: %s: hostname %s verification failed: %s",
				 myname, state->addr, state->name,
				 GAI_STRERROR(error));
			REJECT_PEER_NAME(state);
			break;
		    }
		    if (strcmp(state->addr, rabuf) == 0)
			break;	    /* keep peer name */
		}
		if (res == NULL) {
		    msg_warn("%s: %s: address not listed for hostname %s",
			     myname, state->addr, state->name);
		    REJECT_PEER_NAME(state);
		}
	    }
	    if (res0)
		freeaddrinfo(res0);
#else
	    hp = gethostbyname(state->name);	/* clobbers hp->name!! */
	    if (hp == 0) {
		msg_warn("%s: hostname %s verification failed: %s",
			 state->addr, state->name, HSTRERROR(h_errno));
		REJECT_PEER_NAME(state);
	    } else if (hp->h_length != sizeof(*in)) {
		msg_warn("%s: hostname %s verification failed: bad address size %d",
			 state->addr, state->name, hp->h_length);
		REJECT_PEER_NAME(state);
	    } else {
		int i;
		for (i = 0; /* void */ ; i++) {
		    if (hp->h_addr_list[i] == 0) {
			msg_warn("%s: address not listed for hostname %s",
				 state->addr, state->name);
			REJECT_PEER_NAME(state);
			break;
		    }
		    if (memcmp(hp->h_addr_list[i], (char *)in,
			       sizeof(*in)) == 0)
			break;			/* keep peer name */
		}
	    }
#endif
	}
    }

    /*
     * If it's not Internet, assume the client is local, and avoid using the
     * naming service because that can hang when the machine is disconnected.
     */
    else {
	state->name = mystrdup("localhost");
	state->addr = mystrdup("127.0.0.1");	/* XXX bogus. */
    }

    /*
     * Do the name[addr] formatting for pretty reports.
     */
    state->namaddr =
	concatenate(state->name, "[", state->addr, "]", (char *) 0);
}

/* qmqpd_peer_reset - destroy peer information */

void    qmqpd_peer_reset(QMQPD_STATE *state)
{
    myfree(state->name);
    myfree(state->addr);
    myfree(state->namaddr);
}
