/*++
/* NAME
/*	smtp_connect 3
/* SUMMARY
/*	connect to SMTP server
/* SYNOPSIS
/*	#include "smtp.h"
/*
/*	SMTP_SESSION *smtp_connect(destination, why)
/*	char	*destination;
/*	VSTRING	*why;
/* AUXILIARY FUNCTIONS
/*	SMTP_SESSION *smtp_connect_domain(name, port, why)
/*	char	*name;
/*	unsigned port;
/*	VSTRING	*why;
/*
/*	SMTP_SESSION *smtp_connect_host(name, port, why)
/*	char	*name;
/*	unsigned port;
/*	VSTRING	*why;
/* DESCRIPTION
/*	This module implements SMTP connection management.
/*
/*	smtp_connect() attempts to establish an SMTP session with a host
/*	that represents the named domain.
/*
/*	No session and an smtp_errno of SMTP_OK means that the local
/*	machine is the best mail exchanger for the specified destination.
/*	It is left up to the caller to decide if this is a mailer loop
/*	or if this is a "do what I mean" request.
/*
/*	The destination is either a host (or domain) name or a numeric
/*	address. Symbolic or numeric service port information may be
/*	appended, separated by a colon (":").
/*
/*	By default, the Internet domain name service is queried for mail
/*	exchanger hosts. Quote the domain name with `[' and `]' to
/*	suppress mail exchanger lookups.
/*
/*	Numerical address information should always be quoted with `[]'.
/*
/*	smtp_connect_domain() attempts to make an SMTP connection to
/*	the named host or domain and network port (network byte order).
/*	\fIname\fR is used to look up mail exchanger information via
/*	the Internet domain name system (DNS).
/*	When no mail exchanger is listed for \fIname\fR, the request
/*	is passed to smtp_connect_host().
/*	Otherwise, mail exchanger hosts are tried in order of preference,
/*	until one is found that responds. In order to avoid mailer loops,
/*	the search for mail exchanger hosts stops when a host is found
/*	that has the same preference as the sending machine.
/*
/*	smtp_connect_host() makes an SMTP connection without looking up
/*	mail exchanger information.  The host can be specified as an
/*	Internet network address or as a symbolic host name.
/* DIAGNOSTICS
/*	All routines either return an SMTP_SESSION pointer, or
/*	return a null pointer and set the \fIsmtp_errno\fR
/*	global variable accordingly:
/* .IP SMTP_RETRY
/*	The connection attempt failed, but should be retried later.
/* .IP SMTP_FAIL
/*	The connection attempt failed.
/* .PP
/*	In addition, a textual description of the error is made available
/*	via the \fIwhy\fR argument.
/* SEE ALSO
/*	smtp_proto(3) SMTP client protocol
/* LICENSE
/* .ad
/* .fi
/*	The Secure Mailer license must be distributed with this software.
/* AUTHOR(S)
/*	Wietse Venema
/*	IBM T.J. Watson Research
/*	P.O. Box 704
/*	Yorktown Heights, NY 10598, USA
/*--*/

/* System library. */

#include <sys_defs.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <netdb.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>

#ifdef STRCASECMP_IN_STRINGS_H
#include <strings.h>
#endif

#ifndef INADDR_NONE
#define INADDR_NONE 0xffffff
#endif

/* Utility library. */

#include <msg.h>
#include <vstream.h>
#include <vstring.h>
#include <split_at.h>
#include <mymalloc.h>
#include <inet_addr_list.h>
#include <iostuff.h>
#include <timed_connect.h>
#include <stringops.h>
#include <host_port.h>
#include <sane_connect.h>

/* Global library. */

#include <mail_params.h>
#include <own_inet_addr.h>
#include <pfixtls.h>

/* DNS library. */

#include <dns.h>

#ifdef INET6
#define GAI_STRERROR(error) \
	((error == EAI_SYSTEM) ? strerror(errno) : gai_strerror(error))
#endif
	
/* Application-specific. */

#include "smtp.h"
#include "smtp_addr.h"

/* smtp_force_bind: bind() address */

static void smtp_force_bind(const char *bind_addr,
			    const char *bind_var,
			    int sock,
			    int af)
{
    /*
     * If the bind() call fails, this is considered a non-fatal error.
     * All address conversion errors are fatal.
     */
    char   *myname = "smtp_force_bind";
#ifdef INET6
    char    hbuf[NI_MAXHOST];
    int     aierr;
    struct addrinfo hints, *res;

    memset(&hints, 0, sizeof(hints));
    hints.ai_family = af;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags = AI_PASSIVE | AI_NUMERICHOST;
    snprintf(hbuf, sizeof(hbuf), "%s", bind_addr);
    aierr = getaddrinfo(hbuf, NULL, &hints, &res);
    if (aierr == EAI_NONAME)
	msg_fatal("%s: bad %s parameter: \"%s\"",
		  myname, bind_var, bind_addr);
    if (aierr != 0) {
	if (msg_verbose)
	    msg_warn("%s: getaddrinfo(%s): %s",
		     myname, hbuf, GAI_STRERROR(aierr));
	return;
    }
    aierr = getnameinfo(res->ai_addr, res->ai_addrlen, hbuf, sizeof(hbuf),
			NULL, 0, NI_NUMERICHOST | NI_WITHSCOPEID);
    if (aierr != 0) {
	msg_warn("%s: getnameinfo(): %s",
		 myname, GAI_STRERROR(aierr));
	freeaddrinfo(res);
	return;
    }
    if (bind(sock, res->ai_addr, res->ai_addrlen) < 0)
	msg_warn("%s: bind %s: %m", myname, hbuf);
    else if (msg_verbose)
	msg_info("%s: bind %s", myname, hbuf);
    freeaddrinfo(res);
#else /* INET6 */
    struct sockaddr_in sin;

    memset(&sin, 0, sizeof(sin));
    sin.sin_family = AF_INET;
#ifdef HAS_SA_LEN
    sin.sin_len = sizeof(sin);
#endif
    sin.sin_addr.s_addr = inet_addr(bind_addr);
    if (sin.sin_addr.s_addr == INADDR_NONE) {
	msg_fatal("%s: bad %s parameter: \"%s\"",
		  myname, bind_var, bind_addr);
	return;
    }
    if (bind(sock, (struct sockaddr *)&sin, sizeof(sin)) < 0)
	msg_warn("%s: bind %s: %m", myname, inet_ntoa(sin.sin_addr));
    else if (msg_verbose)
	msg_info("%s: bind %s", myname, inet_ntoa(sin.sin_addr));
#endif /* INET6 */
}

/* smtp_virtual_bind - bind() when acting as virtual host */

static void smtp_virtual_bind(int sock, int af)
{
    char    *myname = "smtp_virtual_bind";
    INET_ADDR_LIST *addr_list;
    int     count;

#ifdef INET6
    int     i, pos;
    char    hbuf[NI_MAXHOST];
    int     aierr;
    struct sockaddr *sa;
    struct addrinfo hints, *loopback = NULL, *res = NULL;

    /*
     * Check whether we are acting as a virtual host
     */
    count = 0;
    pos = 0;
    addr_list = own_inet_addr_list();
    for (i = 0; count < 2 && i < addr_list->used; i++)
	if (((struct sockaddr *)&addr_list->addrs[i])->sa_family == af)
	    count++, pos = i;
    if (count != 1)
	return;

    /*
     * Bind the source address.
     */
    memset(&hints, 0, sizeof(hints));
    hints.ai_family = af;
    hints.ai_socktype = SOCK_STREAM;
    aierr = getaddrinfo(NULL, "0", &hints, &loopback);
    if (aierr != 0) {
	loopback = NULL;
	msg_warn("%s: getaddrinfo(\"0\"): %s",
		 myname, GAI_STRERROR(aierr));
    }

    sa = (struct sockaddr *)&addr_list->addrs[pos];
    aierr = getnameinfo(sa, SA_LEN(sa), hbuf, sizeof(hbuf),
			NULL, 0, NI_NUMERICHOST | NI_WITHSCOPEID);
    if (aierr != 0)
	msg_fatal("%s: getnameinfo() (AF=%d): %s",
		  myname, af, GAI_STRERROR(aierr));

    memset(&hints, 0, sizeof(hints));
    hints.ai_family = af;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags = AI_NUMERICHOST | AI_PASSIVE;
    aierr = getaddrinfo(hbuf, NULL, &hints, &res);
    if (aierr != 0)
	msg_fatal("%s: getaddrinfo(\"%s\"): %s",
		  myname, hbuf, GAI_STRERROR(aierr));

    if (res->ai_addrlen != loopback->ai_addrlen
	|| memcmp(res->ai_addr, loopback->ai_addr, res->ai_addrlen) != 0) {
	if (bind(sock, res->ai_addr, res->ai_addrlen) < 0)
	    msg_warn("%s: bind %s: %m", myname, hbuf);
	else if (msg_verbose)
	    msg_info("%s: bind %s", myname, hbuf);
    } else if (msg_verbose) {
	msg_info("%s: not calling bind(): unusable source "
		 "address from \"%s\"", myname, hbuf);
    }
    if (res)
	freeaddrinfo(res);
    if (loopback)
	freeaddrinfo(loopback);

#else /* INET6 */

    struct sockaddr_in sin;
    unsigned long inaddr;	/*XXX BAD!*/

    /*
     * Check whether we are acting as a virtual host
     */
    addr_list = own_inet_addr_list();
    count = addr_list->used;
    if (count != 1)
	return;

    /*
     * Bind the source address.
     */
    memset(&sin, 0, sizeof(sin));
    sin.sin_family = AF_INET;
#ifdef HAS_SA_LEN
    sin.sin_len = sizeof(sin);
#endif
    memcpy((char *) &sin.sin_addr, addr_list->addrs, sizeof(sin.sin_addr));
    inaddr = (unsigned long)ntohl(sin.sin_addr.s_addr);
    if (!IN_CLASSA(inaddr)
	|| !(((inaddr & IN_CLASSA_NET) >> IN_CLASSA_NSHIFT) == IN_LOOPBACKNET)) {
	if (bind(sock, (struct sockaddr *) & sin, sizeof(sin)) < 0)
	    msg_warn("%s: bind %s: %m", myname, inet_ntoa(sin.sin_addr));
	else if (msg_verbose)
	    msg_info("%s: bind %s", myname, inet_ntoa(sin.sin_addr));
    }
#endif /* INET6 */
}

/* smtp_connect_addr - connect to explicit address */

static SMTP_SESSION *smtp_connect_addr(char *dest, DNS_RR *addr, unsigned port,
				               VSTRING *why)
{
    char   *myname = "smtp_connect_addr";
#ifdef INET6
    struct sockaddr_storage ss;
#else
    struct sockaddr ss;
#endif
    struct sockaddr *sa;
    struct sockaddr_in *sin;
#ifdef INET6
    struct sockaddr_in6 *sin6;
#endif
    SOCKADDR_SIZE salen;
#ifdef INET6
    char hbuf[NI_MAXHOST];
#else
    char hbuf[sizeof("255.255.255.255") + 1];
#endif
    int     sock = -1;
    INET_ADDR_LIST *addr_list;
    int     conn_stat;
    int     saved_errno;
    VSTREAM *stream;
    int     ch;
    char    *bind_addr;
    char    *bind_var;
#ifdef INET6
    char    *addr6_ptr = NULL;
#endif

    sa = (struct sockaddr *)&ss;
    sin = (struct sockaddr_in *)&ss;
#ifdef INET6
    sin6 = (struct sockaddr_in6 *)&ss;
#endif

    /*
     * Sanity checks.
     */
#ifdef INET6
    if (((addr->type==T_A) && (addr->data_len > sizeof(sin->sin_addr))) ||
	((addr->type==T_AAAA) && (addr->data_len > sizeof(sin6->sin6_addr))))
#else
    if (addr->data_len > sizeof(sin->sin_addr))
#endif
    {
	msg_warn("%s: skip address with length %d", myname, addr->data_len);
	smtp_errno = SMTP_RETRY;
	return (0);
    }

    /*
     * Initialize.
     */
    switch (addr->type) {
#ifdef INET6
    case T_AAAA:
	bind_addr = "";
	bind_var = VAR_SMTP_BIND_ADDR6;
	if (*var_smtp_bind_addr6) {
	    addr6_ptr = mystrdup(var_smtp_bind_addr6);
	    if (*addr6_ptr == '[' && addr6_ptr[strlen(addr6_ptr) - 1] == ']') {
		addr6_ptr[strlen(addr6_ptr) - 1] = 0;
		bind_addr = addr6_ptr + 1;
	    } else {
		msg_warn("%s: skip incorrectly bracketed IPv6 address in %s",
		    myname, VAR_SMTP_BIND_ADDR6);
	    }
	}
	memset(sin6, 0, sizeof(*sin6));
	sin6->sin6_family = AF_INET6;
	salen = sizeof(*sin6);
	break;
#endif
    default: /* T_A: */
	bind_addr = var_smtp_bind_addr;
	bind_var = VAR_SMTP_BIND_ADDR;
	memset(sin, 0, sizeof(*sin));
	sin->sin_family = AF_INET;
	salen = sizeof(*sin);
	break;
    }
#ifdef HAS_SA_LEN
    sa->sa_len = salen;
#endif
    if ((sock = socket(sa->sa_family, SOCK_STREAM, 0)) < 0) {
	msg_warn("%s: socket: %m", myname);
	smtp_errno = SMTP_RETRY;
	return (0);
    }
		    
    /*
     * Allow the sysadmin to specify the source address, for example, as "-o
     * smtp_bind_address=x.x.x.x" in the master.cf file.
     */
    if (bind_addr && *bind_addr) {
	smtp_force_bind(bind_addr, bind_var, sock, sa->sa_family);
#ifdef INET6
	if (addr6_ptr)
		myfree(addr6_ptr);
#endif
    } else {
	/*
	 * When running as a virtual host, bind to the virtual interface so that
	 * the mail appears to come from the "right" machine address.
	 */
	smtp_virtual_bind(sock, sa->sa_family);
    }

    /*
     * Connect to the SMTP server.
     */
    switch (addr->type) {
#ifdef INET6
    case T_AAAA:
	/* XXX scope unfriendly */
	memset(sin6, 0, sizeof(*sin6));
	sin6->sin6_port = port;
	sin6->sin6_family = AF_INET6;
	salen = sizeof(*sin6);
	memcpy(&sin6->sin6_addr, addr->data, sizeof(sin6->sin6_addr));
	inet_ntop(AF_INET6, &sin6->sin6_addr, hbuf, sizeof(hbuf));
	break;
#endif
    default: /* T_A */
	memset(sin, 0, sizeof(*sin));
	sin->sin_port = port;
	sin->sin_family = AF_INET;
	salen = sizeof(*sin);
	memcpy(&sin->sin_addr, addr->data, sizeof(sin->sin_addr));
	inet_ntop(AF_INET, &sin->sin_addr, hbuf, sizeof(hbuf));
	break;
    }
#ifdef HAS_SA_LEN
    sa->sa_len = salen;
#endif

    if (msg_verbose)
	msg_info("%s: trying: %s[%s] port %d...",
		 myname, addr->name, hbuf, ntohs(port));
    if (var_smtp_conn_tmout > 0) {
	non_blocking(sock, NON_BLOCKING);
	conn_stat = timed_connect(sock, sa, salen, var_smtp_conn_tmout);
	saved_errno = errno;
	non_blocking(sock, BLOCKING);
	errno = saved_errno;
    } else {
	conn_stat = sane_connect(sock, sa, salen);
    }
    if (conn_stat < 0) {
	vstring_sprintf(why, "connect to %s[%s]: %m",
			addr->name, hbuf);
	smtp_errno = SMTP_RETRY;
	close(sock);
	return (0);
    }

    /*
     * Skip this host if it takes no action within some time limit.
     */
    if (read_wait(sock, var_smtp_helo_tmout) < 0) {
	vstring_sprintf(why, "connect to %s [%s]: read timeout",
			addr->name, hbuf);
	smtp_errno = SMTP_RETRY;
	close(sock);
	return (0);
    }

    /*
     * Skip this host if it disconnects without talking to us.
     */
    stream = vstream_fdopen(sock, O_RDWR);
    if ((ch = VSTREAM_GETC(stream)) == VSTREAM_EOF) {
	vstring_sprintf(why, "connect to %s [%s]: server dropped connection",
			addr->name, hbuf);
	smtp_errno = SMTP_RETRY;
	vstream_fclose(stream);
	return (0);
    }
    vstream_ungetc(stream, ch);

    /*
     * Skip this host if it sends a 4xx greeting.
     */
    if (ch == '4' && var_smtp_skip_4xx_greeting) {
	vstring_sprintf(why, "connect to %s[%s]: server refused mail service",
			addr->name, hbuf);
	smtp_errno = SMTP_RETRY;
	vstream_fclose(stream);
	return (0);
    }

    /*
     * Skip this host if it sends a 5xx greeting.
     */
    if (ch == '5' && var_smtp_skip_5xx_greeting) {
	vstring_sprintf(why, "connect to %s[%s]: server refused mail service",
			addr->name, hbuf);
	smtp_errno = SMTP_RETRY;
	vstream_fclose(stream);
	return (0);
    }
    return (smtp_session_alloc(dest, stream, addr->name, hbuf));
}

/* smtp_connect_host - direct connection to host */

SMTP_SESSION *smtp_connect_host(char *host, unsigned port, VSTRING *why)
{
    SMTP_SESSION *session = 0;
    DNS_RR *addr_list;
    DNS_RR *addr;
    
    /*
     * Try each address in the specified order until we find one that works.
     * The addresses belong to the same A record, so we have no information
     * on what address is "best".
     */
    addr_list = smtp_host_addr(host, why);
    for (addr = addr_list; addr; addr = addr->next) {
	if ((session = smtp_connect_addr(host, addr, port, why)) != 0) {
	    session->best = 1;
	    break;
	}
	msg_info("%s (port %d)", vstring_str(why), ntohs(port));
    }
    dns_rr_free(addr_list);
    return (session);
}

/* smtp_connect_domain - connect to smtp server for domain */

SMTP_SESSION *smtp_connect_domain(char *name, unsigned port, VSTRING *why,
				          int *found_myself)
{
    SMTP_SESSION *session = 0;
    DNS_RR *addr_list;
    DNS_RR *addr;

    /*
     * Try each mail exchanger in order of preference until we find one that
     * responds.  Once we find a server that responds we never try
     * alternative mail exchangers. The benefit of this is that we will use
     * backup hosts only when we are unable to reach the primary MX host. If
     * the primary MX host is reachable but does not want to receive our
     * mail, there is no point in trying the backup hosts.
     */
    addr_list = smtp_domain_addr(name, why, found_myself);
    for (addr = addr_list; addr; addr = addr->next) {
	if ((session = smtp_connect_addr(name, addr, port, why)) != 0) {
	    session->best = (addr->pref == addr_list->pref);
	    break;
	}
	msg_info("%s (port %d)", vstring_str(why), ntohs(port));
    }
    dns_rr_free(addr_list);
    return (session);
}

/* smtp_parse_destination - parse destination */

static char *smtp_parse_destination(char *destination, char *def_service,
				            char **hostp, unsigned *portp)
{
    char   *buf = mystrdup(destination);
    char   *service;
    struct servent *sp;
    char   *protocol = "tcp";		/* XXX configurable? */
    unsigned port;
    const char *err;

    if (msg_verbose)
	msg_info("smtp_parse_destination: %s %s", destination, def_service);

    /*
     * Parse the host/port information. We're working with a copy of the
     * destination argument so the parsing can be destructive.
     */
    if ((err = host_port(buf, hostp, &service, def_service)) != 0)
	msg_fatal("%s in SMTP server description: %s", err, destination);

    /*
     * Convert service to port number, network byte order.
     */
    if (alldig(service) && (port = atoi(service)) != 0) {
	*portp = htons(port);
    } else {
	if ((sp = getservbyname(service, protocol)) == 0)
	    msg_fatal("unknown service: %s/%s", service, protocol);
	*portp = sp->s_port;
    }

    return (buf);
}

/* smtp_connect - establish SMTP connection */

SMTP_SESSION *smtp_connect(char *destination, VSTRING *why)
{
    SMTP_SESSION *session = 0;
    char   *dest_buf = 0;
    char   *host;
    unsigned port;
    char   *def_service = "smtp";	/* XXX configurable? */
    ARGV   *sites;
    char   *dest;
    char  **cpp;
    int     found_myself = 0;

    /*
     * First try to deliver to the indicated destination, then try to deliver
     * to the optional fall-back relays.
     */
    sites = argv_alloc(1);
    argv_add(sites, destination, (char *) 0);
    argv_split_append(sites, var_fallback_relay, ", \t\r\n");

    for (cpp = sites->argv; (dest = *cpp) != 0; cpp++) {

	/*
	 * Parse the destination. Default is to use the SMTP port.
	 */
	dest_buf = smtp_parse_destination(dest, def_service, &host, &port);

	/*
	 * Connect to an SMTP server. Skip mail exchanger lookups when a
	 * quoted host is specified, or when DNS lookups are disabled.
	 */
	if (msg_verbose)
	    msg_info("connecting to %s port %d", host, ntohs(port));
	if (var_disable_dns || *dest == '[') {
	    session = smtp_connect_host(host, port, why);
	} else {
	    session = smtp_connect_domain(host, port, why, &found_myself);
	}
	myfree(dest_buf);

	/*
	 * Done if we have a session, or if we have no session and this host
	 * is the best MX relay for the destination. Agreed, an errno of OK
	 * after failure is a weird way to reporting progress.
	 */
	if (session != 0 || smtp_errno == SMTP_OK || found_myself)
	    break;
    }

    /*
     * Sanity check. The destination must not be empty or all blanks.
     */
    if (session == 0 && dest_buf == 0)
	msg_panic("null destination: \"%s\"", destination);

    /*
     * Pay attention to what could be configuration problems, and pretend
     * that these are recoverable rather than bouncing the mail.
     */
    if (session == 0 && smtp_errno == SMTP_FAIL) {
	if (strcmp(destination, var_relayhost) == 0) {
	    msg_warn("%s configuration problem: %s",
		     VAR_RELAYHOST, var_relayhost);
	    smtp_errno = SMTP_RETRY;
	}
	if (cpp > sites->argv && sites->argc > 1) {
	    msg_warn("%s problem: %s",
		     VAR_FALLBACK_RELAY, var_fallback_relay);
	    smtp_errno = SMTP_RETRY;
	}
    }

    /*
     * Cleanup.
     */
    argv_free(sites);
    return (session);
}
